# Import modules

import pandas as pd
import seaborn as sns
import numpy as np
import matplotlib.pyplot as plt
import math
from scipy import stats
from sklearn.metrics import mean_squared_error, r2_score
from collections import OrderedDict
import textwrap

## Analysis functions

#This function outputs a dataframe with the intraplate variability
# the input columns defines how the data should be grouped.
def calculateIntraplateVariability(df, columns):

    df2 = df.drop(['A425',"A620"], axis =1) #Remove raw data columns, as it is not necessary

    #Calculate mean grouped by `columns` (per sample, per plate, per device, but other conditions for Normalise and dilute)
    means= df2.groupby(columns).mean()
    means = means.rename(columns={"Absorbance (corrected)": "Mean Absorbance (corrected)"})

    #Calculate standard deviation
    stds= df2.groupby(columns).std()
    stds = stds.rename(columns={"Absorbance (corrected)": "Intraplate SD"})
    
    #Concatenate
    stats_df = pd.concat([means, stds], axis=1, sort=False)
    stats_df=stats_df.reset_index()
    stats_df["Intraplate CV (%)"]=100*stats_df["Intraplate SD"]/stats_df["Mean Absorbance (corrected)"]

    return stats_df

#This function outputs a dataframe with the intraplate variability
# the input columns defines how the data should be grouped.
def calculateInterplateVariability(df, columns):
    
    df2 = df.drop(["Intraplate SD", "Intraplate CV (%)"], axis =1)

    #Calculate mean per sample per plate per device
    means= df2.groupby(columns).mean()
    means = means.drop(["Plate"], axis =1)
    means = means.rename(columns={"Mean Absorbance (corrected)": "Mean Interplate Absorbance (corrected)"})
    
    #Calculate standard deviation
    stds= df2.groupby(columns).std()
    stds = stds.drop(["Plate"], axis =1)
    stds = stds.rename(columns={"Mean Absorbance (corrected)": "Interplate SD"})
    
    #Concatenate
    stats_df = pd.concat([means, stds], axis=1, sort=False)
    stats_df=stats_df.reset_index()
    stats_df["Interplate CV (%)"]=100*stats_df["Interplate SD"]/stats_df["Mean Interplate Absorbance (corrected)"]

    return stats_df

## Plotting functions

#This function adds the text defined in `values` to the graph above each group of datapoints, as indicated by the x-tick and its mean value.
def annotate_graph (col_order,col_sort,df,chart, values,annotation_offset):
        
        means_graph = df.sort_values(by=col_sort)
        means_graph = means_graph.groupby(col_sort)['Absorbance (corrected)'].mean()
 
        for ax, device in zip(chart.axes.flat, col_order):
        
            y_coord = means_graph.loc[device]
            
            for xtick, value, yy in zip(ax.get_xticks(),values,y_coord):
            
                ax.text(xtick, yy+annotation_offset,value, fontsize=10, horizontalalignment='center') #add text)

                
##Add legend to plots in Absorbance plots
def addLegendToAllPlots(chart, position):
    
    for ax in chart.axes.flat:
            
        ax.legend(title = hue_abs,             
                  loc=position, 
                  #bbox_to_anchor=(1.2, 0.6), 
                  ncol=1,
                  frameon=False,
                  title_fontsize = 12)

## Different experiments will have data sets with different columns which require different plotting. 
## This function sets all the variables needed so that the same code can be used to do the analysis and make graphs for the different datasets.

def setElementParameters(element):
    #Set columns based on element dataset, for grouping
    if element == MakeMixtures:

        file = "/data_MakeMixtures.csv"
        columns_intra = ["Sample Name", "Plate", "Device"] 
        columns_inter = ["Sample Name", "Device"] 

        #Figure variables
        columns_reorder_raw = ["Well","A425","A620","Absorbance (corrected)","Sample Name","Plate","Device"]
        x_cat = "Sample Name"
        ylim = [0,3]
        row_abs = None
        col_abs = "Device"
        hue_abs = "Plate"

        col_intra = None
        col_order = None
        hue_intra = "Plate"

        hue_inter = None
        hue_order_inter = None


    elif element == Aliquot:

        file = "/data_Aliquot.csv"
        columns_intra = ["Volume /uL", "Plate", "Device"] 
        columns_inter = ["Volume /uL", "Device"] 

        #Figure variables
        columns_reorder_raw = ["Well","A425","A620","Absorbance (corrected)","Volume /uL","Plate","Device"]
        x_cat = "Volume /uL"
        ylim = [0,1.8]
        row_abs = None
        col_abs = "Device"
        hue_abs = "Plate"

        col_intra = None
        col_order = None
        hue_intra = "Plate"

        hue_inter = None
        hue_order_inter = None



    elif element == MixOnto:

        file = "/data_MixOnto.csv"
        columns_intra = ["Mix Name", "Plate", "Device"] 
        columns_inter = ["Mix Name", "Device"] 


        #Figure variables
        columns_reorder_raw = ["Well","A425","A620","Absorbance (corrected)","Mix Name","Plate","Device"]
        x_cat = "Mix Name"
        ylim = [0,1]
        row_abs = None
        col_abs = "Device"
        hue_abs = "Plate"

        col_intra = None
        col_order = None
        hue_intra = "Plate"

        hue_inter = None
        hue_order_inter = None


    elif element == Dilute:

        file = "/data_Dilute.csv"
        columns_intra = ["Concentration /uM", "Plate", "Device"] 
        columns_inter = ["Concentration /uM", "Device"] 

        #Figure variables
        columns_reorder_raw = ["Well","A425","A620","Absorbance (corrected)","Sample Name","Concentration /uM","Plate","Device"]
        x_cat = "Concentration /uM"
        ylim = [0,3]
        row_abs = "Plate"
        col_abs = "Device"
        hue_abs = "Device"

        col_intra = None
        col_order = None
        hue_intra = "Plate"

        hue_inter = None
        hue_order_inter = None



    elif element == Normalize:

        file = "/data_Normalize.csv"    
        columns_intra = ["Sample Name", "Plate", "Device","Normalization"] 
        columns_inter = ["Sample Name", "Device","Normalization"]

        #Figure variables
        columns_reorder_raw = ["Well","A425","A620","Absorbance (corrected)","Sample Name","Plate","Device","Normalization"]
        x_cat = "Sample Name"
        ylim = [0.5,1.6]
        row_abs = "Device"
        col_abs = "Normalization"
        hue_abs = "Plate"

        col_intra = "Normalization"
        col_order = ['Without', 'With']
        hue_intra = "Plate"

        hue_inter = "Normalization"
        hue_order_inter = ['Without', 'With']
        
        
    return file, columns_intra, columns_inter, columns_reorder_raw, x_cat, ylim, row_abs, col_abs, hue_abs, col_intra, col_order, hue_intra, hue_inter, hue_order_inter

## Global Figure Settings

sns.set(rc={'figure.figsize':(25, 20)})

sns.set_style("ticks")
              #{'font.family':'sans-serif', 'font.sans-serif':'Roboto'})
sns.set_context("paper", 
                font_scale=1.2, 
                rc={
                    #Font sizes
                    'font.size': 12,
                    'axes.labelsize': 12,
                    'axes.titlesize': 12,
                    'legend.fontsize': 12,
                    'legend.title_fontsize': 12,
                    'xtick.labelsize': 12.0,
                    'ytick.labelsize': 12,                 
                    #Line thickeness
                    'axes.linewidth': 1,
                    'grid.linewidth': 1,
                    'lines.linewidth': 1                    
                    #'lines.markeredgewidth': 1
                    #'lines.markersize': 11.2,
                    #'patch.linewidth': 0.48,
                    #'xtick.major.pad': 11.2,
                    #'xtick.major.width': 1.6,
                    #'xtick.minor.width': 0.8,
                    #'ytick.major.pad': 11.2,
                    #'ytick.major.width': 1.6,
                    #'ytick.minor.width': 0.8
                   }
               )

# Create an array with the colors you want to use
colors = ["#5E9DFF", # Primary 30
          "#FFD375", # BrandMark 30
          "#D07EFF"] # Secondary 30

#Color for marker edge in scatter
edgeMarkerColour = "#0D0D0D" # Neutral 90

# Set your custom color palette
sns.set_palette(sns.color_palette(colors))


#Save figure function
def savefigure(chart,pathToSaveFigs,fileName):
    return chart.savefig(pathToSaveFigs+fileName, dpi=900)


#Location of raw data file
#pathToData='C:/Users/Nuno Leitao/Dropbox (Synthace Limited)/Nunos docs/Lab projects/Core Elements - Technical demonstration/'
pathToData='G:/Shared drives/Synthace/Product/Element Content/Technical Documents/'

##List of elements
Aliquot = "Aliquot"
MakeMixtures = "Make Mixtures"
MixOnto = "Mix Onto"
Normalize = "Normalize"
Dilute = "Dilute"

elements = [Aliquot, MakeMixtures, MixOnto, Normalize, Dilute]

# Iterate through each element and:
# 1. get raw data
# 2. correct absorbance
# 3. calculate intra and interplate variance
# 4. round values
# 5. save statistics to csv
# 6. plot corrected absorbance
# 7. plot intraplate CV
# 8. plot interplate CV

for element in elements:
    
    file, columns_intra, columns_inter, columns_reorder_raw, x_cat, ylim, row_abs, col_abs, hue_abs, col_intra, col_order, hue_intra, hue_inter, hue_order_inter = setElementParameters(element)

    ######################################################################
    ######################## DATA HANDLING ###############################
    ######################################################################
    
    
    #Destination of generated figures
    pathToSaveFigs = pathToData+"Figures/"

    #Read CSV
    data = pd.read_csv(pathToData+'Structured data files/'+file)

    #Subtract A620 from A425
    data["Absorbance (corrected)"]=data["A425"]-data["A620"]
    #Reorder columns
    data=data[columns_reorder_raw]

    ######################################################################
    ############################ ANALYSIS ################################
    ######################################################################    
    
    #Calculate intraplate variation
    stats_intra = calculateIntraplateVariability(data, columns_intra)
    
    #Calculate interplate variation
    stats_inter = calculateInterplateVariability(stats_intra, columns_inter) #Interplate variation calculated from intraplate variation

    #Round values before graphing
    stats_intra=stats_intra.round(2)
    stats_inter=stats_inter.round(2)
    
    #Save values to a csv
    stats_intra.to_csv(pathToData+"stats_intra"+element+".csv", index=False)
    stats_inter.to_csv(pathToData+"stats_inter"+element+".csv", index=False)

    ##Generates summary statistics associated with linear regression for Dilute element - Does not separate by sample name
    if element == Dilute:

        stats_intra.drop(stats_intra[stats_intra["Concentration /uM"] == 0 ].index, inplace=True)
        stats_inter.drop(stats_inter[stats_inter["Concentration /uM"] == 0 ].index, inplace=True)

        devices = data["Device"].unique().tolist()
        plates = data["Plate"].unique().tolist()
        samples = data["Sample Name"].unique().tolist()

        summary_stats = pd.DataFrame(columns=["Device","Plate","Slope","Intercept","R2_value","P_value","Std_err"])

        for plate in plates:

            for device in devices:

                slope, intercept, r_value, p_value, std_err = stats.linregress(data[(data['Device'] == device) & (data['Plate'] == plate)]["Concentration /uM"], data[(data['Device'] == device) & (data['Plate'] == plate)]['Absorbance (corrected)'])
                summary_stats=summary_stats.append(
                        {
                        'Device': device,
                        'Plate': plate,
                        'Slope': slope,
                        'Intercept': intercept,
                        'R2_value': r_value*r_value,
                        'P_value': p_value,
                        'Std_err': std_err,
                        },ignore_index=True)

        summary_stats.to_csv(pathToData+"Stats_by_sample.csv", index=False)
    
    ######################################################################
    ############################ PLOTTING ################################
    ######################################################################
    
    ##################### ABSORBANCE OVER SAMPLE #########################

    
    # Plot corrected absorbance over sample name, per plate, per device
    if element == Dilute: #If Dilute, plot the dilution curves

        chart = sns.lmplot(    data=data,
                                x="Concentration /uM",
                                y='Absorbance (corrected)',
                                hue="Device",
                                col = "Device",
                                row = "Plate",
                                fit_reg = True,
                                truncate= True,
                                aspect = 1.2,
                                height = 4,
                                line_kws={'ls':'dashdot',"color":"black","lw":0.75},
                                scatter_kws={"s":50,"edgecolor":edgeMarkerColour,"linewidth":0.5}
                          )

        # Annotate r^2 and p value on each plot.
        for ax, r_value, p_value in zip(chart.axes.flat, summary_stats["R2_value"], summary_stats["P_value"]):     
            ax.text(0.05, 
                    0.95,
                   # 'r\u00b2 = {:.3f}\np = {:.3g}'.format(r_value, p_value), #real p value
                    'r\u00b2 = {:.3f}\np < 0.001'.format(r_value), 
                    fontsize=12,
                    horizontalalignment='left',
                    verticalalignment='top',
                    transform = ax.transAxes)

        #Set y-axis limits
        chart.set(ylim=[-0.2,1.2])
        chart.set(xlim=[-10,160])

        #Modify column title
        chart.set_titles(col_template = "{col_name}")

        ## Add borders
        sns.despine(bottom = False, left = False, right = False, top = False)

    else: # for all other elements, plot a categorical plot showing Absorbance over sample/mix name

        chart = sns.catplot(
                                data=data,
                                x=x_cat,
                                y='Absorbance (corrected)',
                                dodge=True,
                                col = col_abs,
                                row = row_abs,
                                hue = hue_abs,
                                margin_titles=True,
                                s=8,
                                edgecolor = edgeMarkerColour,
                                legend_out=True,
                                legend=False,
                                linewidth = 0.5,
                                aspect = 1.2,
                                height = 4,
                                kind="strip",
                                jitter=0.3
                                #split=True
                                )

        #Set y-axis limits
        chart.set(ylim=ylim)

        #Add annotations
        # Annotation vary across elements
        if element == Aliquot:

            #Annotate datapoints with volume values
            devices = data["Device"].unique().tolist()
            values = ["75 uL","100 uL","150 uL","200 uL"]
            offset = 0.15
            annotate_graph(devices,columns_inter[::-1], data,chart, values, offset)

            #Add legend
            addLegendToAllPlots(chart, "upper left")

            #Modify plot title
            for ax,device in zip(chart.axes.flat, devices):
                ax.set_title(device)

        elif element == MakeMixtures:

            #Annotate datapoints with concentration values
            devices = data["Device"].unique().tolist()
            values = ["200 uM","167 uM","143 uM","125 uM","111 uM","100 uM"]
            offset = 0.26
            annotate_graph(devices,columns_inter[::-1], data,chart, values,offset)

            #Add legend
            addLegendToAllPlots(chart, "upper right")

            #Modify plot title
            for ax,device in zip(chart.axes.flat, devices):
                ax.set_title(device)

        elif element == MixOnto:

            #Annotate datapoints with concentration values
            devices = data["Device"].unique().tolist()
            values = ["50 uM","25 uM","50 uM","25 uM"]
            offset = 0.1
            annotate_graph(devices,columns_inter[::-1], data,chart, values, offset)

            #Add legend
            addLegendToAllPlots(chart, "upper right")

            #Modify plot title
            for ax,device in zip(chart.axes.flat, devices):
                ax.set_title(device)


        elif element == Normalize:

            #Change default margin titles
            [plt.setp(ax.texts, text="") for ax in chart.axes.flat] # remove the original texts
            chart.set_titles(row_template = '{row_name}')
            chart.axes[0][0].set_title("Without Normalization")
            chart.axes[0][1].set_title("With Normalization")

            #Rotate x labels
            for ax in chart.axes.ravel():
                ax.set_xticklabels(ax.get_xticklabels(), rotation=45,horizontalalignment='right')

            #Annotate each value with its concentration and add figure legend to plots on the right
            devices = ["Gilson Pipetmax","Gilson Pipetmax","Hamilton Microlab STAR","Hamilton Microlab STAR","Tecan Freedom EVO","Tecan Freedom EVO"]
            means_graph = data.sort_values(by=columns_inter[::-1])
            means_graph = means_graph.groupby(columns_inter[::-1])['Absorbance (corrected)'].mean()
            j=0        

            for ax in chart.axes.flat:

                if (j % 2) == 0: #Plots on the left hand side 

                    annotation_offset =0.1
                    values = ["100 uM","70 uM","65 uM","85 uM","80 uM","95 uM","75 uM","90 uM"]
                    y_coord = means_graph.loc["Without",devices[j]]

                    #Annotate values
                    for xtick, value,yy in zip(ax.get_xticks(),values,y_coord):
                        ax.text(xtick, yy+annotation_offset,value, fontsize=10, horizontalalignment='center')

                else: #Plots on the left hand side 
                    annotation_offset =0.1
                    values = ["50 uM","50 uM","50 uM","50 uM","50 uM","50 uM","50 uM","50 uM"]
                    y_coord = means_graph.loc["With","Gilson Pipetmax"][0]

                    for xtick, value in zip(ax.get_xticks(),values):

                        #Annotate values
                        ax.text(xtick, y_coord+annotation_offset,value, fontsize=10, horizontalalignment='center') 

                        #Add figure legend
                        ax.legend(title = hue_abs, 
                                  loc='upper right', 
                                  #bbox_to_anchor=(1.2, 0.6), 
                                  ncol=1,
                                  frameon=False,
                                  title_fontsize = 12)
                j=j+1

        ## Add borders
        sns.despine(bottom = False, left = False, right = False, top = False)

    fileName = element+" - Absorbance over sample, per plate, per device.jpg"
    savefigure(chart,pathToSaveFigs,fileName)
    
    
    ##################### INTRAPLATE CV #########################
    
    
    # Plot intraplate absorbance over sample name, per plate, per device
    chart = sns.catplot(
                                data=stats_intra,
                                x='Device',
                                y='Intraplate CV (%)',
                                dodge=True,
                                col = col_intra,
                                col_order = col_order,
                                hue = hue_intra,
                                margin_titles=True,
                                s=8,
                                legend_out=False,
                                legend=True,
                                linewidth = 0.5,
                                edgecolor = edgeMarkerColour,
                                aspect = 1.4,
                                height = 4,
                                kind="strip",
                                jitter=0.2
                                #split=True
                                )

    #Device names are long and overlap. Make sure they break into a new line
    max_width = 15
    for ax in chart.axes.ravel():
        ax.set_xticklabels(textwrap.fill(x.get_text(), max_width) for x in ax.get_xticklabels())

    #Set y scale limits
    chart.set(ylim=[0,8])

    if element == Normalize:
        chart.axes[0][0].legend(title = hue_intra, edgecolor="white",title_fontsize = 12)
        chart.axes[0][1].legend(title = hue_intra, edgecolor="white",title_fontsize = 12)

        # Add reference value horizontal line
        chart.axes[0][0].axhline(5,ls="dashdot",color="black")
        chart.axes[0][1].axhline(5,ls="dashdot",color="black")

        norm = ["Without Normalization","With Normalization"]
        for ax,nor in zip(chart.axes.flat, norm):
                ax.set_title(nor)

    else:
        chart.axes[0][0].legend(title = hue_intra, edgecolor="white",title_fontsize = 12)

        # Add reference value horizontal line
        chart.axes[0][0].axhline(5,ls="dashdot",color="black")

    ## Add borders
    sns.despine(bottom = False, left = False, right = False, top = False)

    plt.margins(x=0.2)


    #Save figure
    fileName = element+" - Intraplate CV over device, per plate.jpg"
    savefigure(chart,pathToSaveFigs,fileName)
    
    
    ######################### INTERPLATE CV ##############################
    
    
    # Plot intraplate absorbance over sample name, per plate, per device
    chart = sns.catplot(
                            data=stats_inter,
                            x='Device',
                            y='Interplate CV (%)',
                            hue=hue_inter,
                            hue_order = hue_order_inter,
                            margin_titles=True,
                            s=8,
                            legend_out=False,
                            legend=False,
                            linewidth = 0.5,
                            edgecolor = edgeMarkerColour,
                            aspect = 1.4,
                            height = 4,
                            kind="strip",
                            jitter=0.2,
                            dodge = True
                            )

    #Device names are long and overlap. Make sure they break into a new line
    max_width = 15
    for ax in chart.axes.ravel():
        ax.set_xticklabels(textwrap.fill(x.get_text(), max_width) for x in ax.get_xticklabels())

    #Set y scale limits
    chart.set(ylim=[0,5])

    # Add reference value horizontal line
    plt.axhline(y=2.5, ls='dashdot', color= edgeMarkerColour)

    ## Add borders
    sns.despine(bottom = False, left = False, right = False, top = False)

    plt.margins(x=0.1)

    if element == Normalize:
        #Format legend
        chart.axes[0][0].legend(title = hue_inter, edgecolor="white",title_fontsize = 12)

    fileName = element+" - Interplate CV over device, per plate.jpg"
    savefigure(chart,pathToSaveFigs,fileName)
